/* Copyright 2010, Unpublished Work of Technologic Systems
 * All Rights Reserved.
 *
 * THIS WORK IS AN UNPUBLISHED WORK AND CONTAINS CONFIDENTIAL,
 * PROPRIETARY AND TRADE SECRET INFORMATION OF TECHNOLOGIC SYSTEMS.
 * ACCESS TO THIS WORK IS RESTRICTED TO (I) TECHNOLOGIC SYSTEMS 
 * EMPLOYEES WHO HAVE A NEED TO KNOW TO PERFORM TASKS WITHIN THE SCOPE
 * OF THEIR ASSIGNMENTS AND (II) ENTITIES OTHER THAN TECHNOLOGIC
 * SYSTEMS WHO HAVE ENTERED INTO APPROPRIATE LICENSE AGREEMENTS.  NO
 * PART OF THIS WORK MAY BE USED, PRACTICED, PERFORMED, COPIED, 
 * DISTRIBUTED, REVISED, MODIFIED, TRANSLATED, ABRIDGED, CONDENSED, 
 * EXPANDED, COLLECTED, COMPILED, LINKED, RECAST, TRANSFORMED, ADAPTED
 * IN ANY FORM OR BY ANY MEANS, MANUAL, MECHANICAL, CHEMICAL, 
 * ELECTRICAL, ELECTRONIC, OPTICAL, BIOLOGICAL, OR OTHERWISE WITHOUT
 * THE PRIOR WRITTEN PERMISSION AND CONSENT OF TECHNOLOGIC SYSTEMS.
 * ANY USE OR EXPLOITATION OF THIS WORK WITHOUT THE PRIOR WRITTEN
 * CONSENT OF TECHNOLOGIC SYSTEMS COULD SUBJECT THE PERPETRATOR TO
 * CRIMINAL AND CIVIL LIABILITY.
 */

/* System control registers for the TS-4800.  Base address on
 * TS-4800 is 0xb0010000.  Features:
 *   - Watchdog (can be triggered to reset the board)
 *   - 55 DIO lines
 *   - two LEDs
 *   - board control lines
 *   - 72MHz counter
 *   - ID registers and flashrom containing board info
 *
 * Register map:
 * base + 0x0: board id register.  Reads 0x4800 on TS-4800
 * base + 0x2: submodel, fpga revision, DIO and LED control - (R/W)
 *   bit 15: Reset FPGA
 *   bit 14: Strap-Low signal (0 - low)
 *   bit 13: SD Power enable (1 - on)
 *   bit 12: Boot mode
 *   bit 11: baseboard 12MHz clk enable (1 - on)
 *   bit 10: reserved
 *   bit 9: can1_irq enable (1 - can1_irq_en or irq_cascade_dis)
 *   bit 8: ext_temp_bit (1 - ?)
 *   bit 7-4: board submodel (RO) - 0x0 on production TS-4800
 *   bit 3-0: FPGA revision
 * base + 0x4: 32-bit 72MHz free running counter (16 LSBs)
 * base + 0x6: 32-bit 72MHz free running counter (16 MSBs)
 * base + 0x8: 32-bit 1MHz free running counter (16 LSBs)
 * base + 0xa: 32-bit 1MHz free running counter (16 MSBs)
 * base + 0xc: 16-bit random number
 * base + 0xe: Watchdog feed register (write only)
 *   bit 15-2: reserved
 *   write value 0x0: feed watchdog for another .338s (approx)
 *   write value 0x1: feed watchdog for another 2.706s (approx)
 *   write value 0x2: feed watchdog for another 10.824 (approx)
 *   write value 0x3: disable watchdog
 * base + 0x10: TS-4800 control register
 *   bit 15: Green LED control (1 - on)
 *   bit 14: Red LED control (1 - on)
 *   bit 13: IRQ7 enable (1 - on)
 *   bit 12: IRQ6 enable (1 - on)
 *   bit 11: IRQ5 enable (1 - on)
 *   bit 10: reserved
 *   bit 9: Sound Codec enable (1 - on)
 *   bit 8: Motor Controller enable (1 - on)
 *   bit 7: SPI enable (1 - on)
 *   bit 6: TOUCHSCREEN enable (1 - on)
 *   bit 5: XUART5 enable (1 - on)
 *   bit 4: XUART4 enable (1 - on)
 *   bit 3: XUART3 enable (1 - on)
 *   bit 2: XUART0 enable (1 - on)
 *   bit 1: CAN1 enable (1 - on)
 *   bit 0: CAN0 enable (1 - on)
 * base + 0x12: External bus configuration register
 * base + 0x14: Tagmem control (RW)
 *   bit 15-4: reserved
 *   bit 3: Lattice tagmem clock (RW)
 *   bit 2: Lattice tagmem serial-in (RW)
 *   bit 1: Lattice tagmem CSn (RW)
 *   bit 0: Lattice tagmem serial-out (RO)
 * base + 0x16: SPI PLL phase, latched mode bits, scratch reg
 *   bit 15: Input counter1 enable  (1 - on)
 *   bit 14: Input counter0 enable  (1 - on)
 *   bit 13: Macrocontroller RS422 enable (xuart_rx[0] = dio_i[13]) (RW)
 *   bit 12: TS-8510 specific RS422 enable (xuart_rx[4] = dio_i[13]) (RW)
 *   bit 11: reserved
 *   bit 10-6: PLL phase (set by TS-BOOTROM) (RW)
 *   bit 5: mode3 latched bootstrap bit (RO)
 *   bit 4: Reset switch enable (1 - auto reboot when dio_i[9] == 0) (RW)
 *   bit 3-2: scratch reg (RW)
 *   bit 1: mode2 latched bootstrap bit (RO)
 *   bit 0: mode1 latched bootstrap bit (RO)
 * base + 0x1a: 16-bit input counter0
 * base + 0x1c: 16-bit input counter1
 * base + 0x1e: 16-bit test register
 * base + 0x20: DIO input data for DIO 15(MSB)-0(LSB) (RO)
 * base + 0x22: DIO output data for DIO 15(MSB)-0(LSB)
 * base + 0x24: DIO direction for DIO 15(MSB)-0(LSB) (1-output)
 * base + 0x28: DIO input data for DIO 31(MSB)-16(LSB) (RO)
 * base + 0x2a: DIO output data for DIO 31(MSB)-16(LSB)
 * base + 0x2c: DIO direction for DIO 31(MSB)-16(LSB) (1-output)
 * base + 0x30: DIO input data for DIO 47(MSB)-32(LSB) (RO)
 * base + 0x32: DIO output data for DIO 47(MSB)-32(LSB)
 * base + 0x34: DIO direction for DIO 47(MSB)-32(LSB) (1-output)
 * base + 0x38: DIO input data for DIO 55(MSB)-48(LSB) (RO)
 * base + 0x3a: DIO output data for DIO 55(MSB)-48(LSB)
 * base + 0x3c: DIO direction for DIO 55(MSB)-48(LSB) (1-output)
 *
 * * Watchdog by default comes out of reset armed for .338 seconds.  
 *   TS-BOOTROM firmwarm feeds for 10.824 and OS code has 10.824 seconds
 *   to take over.
 *
 * * DIO FPGA Pad Map:
 *   DIO[14:0]: dio_pad[14:0];
 *   DIO[16:15]: reserved
 *   DIO[20:17]: dio_pad[20:17];
 *   DIO[36:21]: mux_ad_pad[15:0];
 *   DIO[41:37]: {bus_wait_padn, bus_bhe_padn, bus_cs_padn, bus_dir_pad, bus_ale_padn};
 *   DIO[45:42]: {spi_mosi_pad, spi_cs_padn, spi_clk_pad, spi_miso_pad};
 *   DIO[47:46]: {can_txd_pad, can_rxd_pad};
 *   DIO[49:48]: {uart0_txd_pad, uart0_rxd_pad};
 *   DIO[51:50]: {uart3_txd_pad, uart3_rxd_pad};
 *   DIO[53:52]: {uart4_txd_pad, uart4_rxd_pad};
 *   DIO[55:54]: {uart5_txd_pad, uart5_rxd_pad};
 *
 */

module syscon(
  wb_clk_i,
  wb_rst_i,

  wb_cyc_i,
  wb_stb_i,
  wb_we_i,
  wb_adr_i,
  wb_dat_i,
  wb_sel_i,
  wb_dat_o,
  wb_ack_o,

  dio_i,
  dio_oe_o,
  dio_o,

  mode1_i,
  mode2_i,
  mode3_i,
  pllphase_o,
  timer_i,
  led_grn_o,
  led_red_o,
  boot_mode_i,
  reset_fpga_o,
  sd_power_o,
  bbclk_o,
  irq_can1_o,
  strap_low_o,
  bus_config_o,
  core_enable_o,
  ext_temp_i,
  en_ts8510_rs422_o,
  en_rs422_o,
  counter_en_o,
  usec_o,
);

parameter p_submodel = 4'h0;
parameter p_pldrev = 4'h6;

input wb_clk_i, wb_rst_i, wb_cyc_i, wb_stb_i, wb_we_i;
input [16:0] wb_adr_i;
input [15:0] wb_dat_i;
input [3:0] wb_sel_i;
output [15:0] wb_dat_o;
output wb_ack_o;

input [55:0] dio_i;
output [55:0] dio_oe_o;
output [55:0] dio_o;

input mode1_i, mode2_i, mode3_i;
output [4:0] pllphase_o;
input [31:0] timer_i;
output led_grn_o, led_red_o;
input boot_mode_i;
output reset_fpga_o;
output sd_power_o;
output irq_can1_o;
output bbclk_o;
output strap_low_o;
output [15:0] bus_config_o;
output [15:0] core_enable_o;
input ext_temp_i;
output en_ts8510_rs422_o;
output en_rs422_o;
output [1:0] counter_en_o;
output [31:0] usec_o;

reg reset_fpga, sd_power, strap_low, irq_can1, bbclk;
assign sd_power_o = sd_power;
assign strap_low_o = strap_low;
assign irq_can1_o = irq_can1;
assign bbclk_o = bbclk;

reg [55:0] dio_oe;
reg [55:0] dio;
assign dio_oe_o = dio_oe;
assign dio_o = dio;

reg [15:0] bus_config;
assign bus_config_o = bus_config;

reg [15:0] core_enable;
assign core_enable_o = core_enable;
wire led_red = core_enable[14];
wire led_grn = core_enable[15];

assign led_red_o = led_red;
assign led_grn_o = led_grn;

reg [15:0] test_reg;

wire iclk;
OSCE #(.NOM_FREQ("3.1")) internal_clk(iclk);
wire random_clk = iclk;
reg [15:0] rndcnt;
reg shift;

reg [1:0] wdogctl;
reg [9:0] wdogcnt;
reg feed_en, reboot;
wire sed_err, feed_bsy, feed_rdreq;
wire [15:0] wdog_dat;

reg [1:0] counter_en;
assign counter_en_o = counter_en;
reg en_ts8510_rs422, en_rs422;
assign en_ts8510_rs422_o = en_ts8510_rs422;
assign en_rs422_o = en_rs422;
reg [4:0] pllphase;
reg [1:0] scratch;
reg resetsw_en;
reg [1:0] resetsw;

assign reset_fpga_o = reset_fpga | reboot | (resetsw_en && (resetsw == 2'b00));

always @(posedge random_clk or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    rndcnt <= 16'd0;
    shift <= 1'b0;
    wdogctl <= 2'd2;
    wdogcnt <= 10'd0;
    resetsw <= 2'b11;
  end else begin
    shift <= 1'b0;
    rndcnt <= rndcnt + 1'b1;
    if (rndcnt == 16'd0) begin
      shift <= 1'b1;
      resetsw[0] <= dio_i[9];
      resetsw[1] <= resetsw[0] | dio_i[9];
      wdogcnt <= wdogcnt + 1'b1;
    end
    if (feed_rdreq) begin
      wdogctl <= wdog_dat;
      wdogcnt <= 10'd0;
    end
  end
end
always @(*) begin
  case (wdogctl)
  2'd0: reboot = wdogcnt[4]; /* approx .338 seconds */
  2'd1: reboot = wdogcnt[7]; /* approx 2.706 seconds */
  2'd2: reboot = wdogcnt[9]; /* approx 10.824 seconds */
  2'd3: reboot = 1'b0;
  endcase
end

wire [15:0] random;
crc randnumgen(
  .dat_i(wb_clk_i),
  .clk_i(random_clk),
  .clken_i(shift),
  .rst_i(wb_rst_i),
  .shift_i(1'b0),
  .crc16_o(random)
);
/*
SEDBA sedcore (
  .SEDENABLE(1'b1),
  .SEDSTART(1'b1),
  .SEDFRCERRN(1'b1),
  .SEDERR(sed_err)
);
*/
resync_fifo watchdogfifo(
  .rst_i(wb_rst_i),
  .wrclk_i(wb_clk_i),
  .wrreq_i(feed_en),
  .data_i(wb_dat_i),
  .full_wrclk_o(feed_bsy),
  .rdclk_i(random_clk),
  .rdreq_i(feed_rdreq),
  .q_o(wdog_dat),
  .full_rdclk_o(feed_rdreq)
);


wire spi_so;
reg spi2_clk, spi2_si, spi2_csn;
tagmem tagmemcore(
  .CLK(spi2_clk),
  .SI(spi2_si),
  .CS(spi2_csn),
  .SO(spi_so)
);


reg [6:0] count;
reg [31:0] usec;
assign usec_o = usec;
always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    count <= 7'd0;
    usec <= 32'd0;
  end else begin
    count <= count + 1'b1;
    if (count == 7'd71) begin
      count <= 7'd0;
      usec <= usec + 1'b1;
    end
  end
end

reg [1:0] counter_wcyc;
reg [15:0] counter0;
wire counter0_clk = dio_i[4];
wire counter0_rst = wb_rst_i || counter_wcyc[0];
always @(negedge counter0_clk or posedge counter0_rst) begin
	if (counter0_rst) counter0 <= 16'h0000;
	else if (counter_en[0]) counter0 <= counter0 + 1'b1;
end
reg [15:0] counter1;
wire counter1_clk = dio_i[6];
wire counter1_rst = wb_rst_i || counter_wcyc[1];
always @(negedge counter1_clk or posedge counter1_rst) begin
	if (counter1_rst) counter1 <= 16'h0000;
	else if (counter_en[1]) counter1 <= counter1 + 1'b1;
end

always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
    dio_oe <= 56'h00000000000000;
    dio <=    56'h00000000000000;
	reset_fpga <= 1'b0;
    strap_low <= 1'b0;
	sd_power <= 1'b1;
	irq_can1 <= 1'b0;
	bbclk <= 1'b0;
	test_reg <= 16'd0;
    bus_config <= 16'd0;
    core_enable <= 16'h8200;
    scratch <= 2'b00;
	//resetsw_en is on by default
    resetsw_en <= 1'b1;
    pllphase <= 5'h18;
    spi2_clk <= 1'b0;
    spi2_si <= 1'b0;
    spi2_csn <= 1'b1;
    en_ts8510_rs422 <= 1'b0;
    en_rs422 <= 1'b0;
	counter_wcyc <= 2'b00;
	counter_en <= 2'b00;
  end
  else begin
	counter_wcyc <= 2'b00;
    if (wb_cyc_i && wb_stb_i && wb_we_i) begin
      case (wb_adr_i[7:0])
      8'h02: {reset_fpga, strap_low, sd_power, bbclk, irq_can1} <= {wb_dat_i[15:13], wb_dat_i[11], wb_dat_i[9]}; 
      8'h10: core_enable <= wb_dat_i;
      8'h12: bus_config <= wb_dat_i;
      8'h14: {spi2_clk, spi2_si, spi2_csn} <= wb_dat_i[3:1];
      8'h16: {counter_en, en_rs422, en_ts8510_rs422, pllphase, resetsw_en, scratch} <=
	    {wb_dat_i[15:14], wb_dat_i[13:12], wb_dat_i[10:6], wb_dat_i[4], wb_dat_i[3:2]};
      8'h1a: counter_wcyc[0] <= 1'b1;
	  8'h1c: counter_wcyc[1] <= 1'b1;
      8'h1e: test_reg[15:0] <= wb_dat_i;
      8'h22: dio[15:0] <= wb_dat_i[15:0];
      8'h24: dio_oe[15:0] <= wb_dat_i[15:0];
      8'h2a: dio[31:16] <= wb_dat_i[15:0];
      8'h2c: dio_oe[31:16] <= wb_dat_i[15:0];
      8'h32: dio[47:32] <= wb_dat_i[15:0];
      8'h34: dio_oe[47:32] <= wb_dat_i[15:0];
      8'h3a: dio[55:48] <= wb_dat_i[7:0];
      8'h3c: dio_oe[55:48] <= wb_dat_i[7:0];
      endcase
    end
  end
end

reg wb_ack;
assign wb_ack_o = wb_ack;
reg [15:0] wb_dat;
assign wb_dat_o = wb_dat;

always @(posedge wb_clk_i or posedge wb_rst_i) begin
  if (wb_rst_i) begin
	wb_ack <= 1'b0;
    wb_dat <= 16'h0000;
    //feed_en <= 1'b0;
  end else begin
    wb_ack <= wb_cyc_i && wb_stb_i && !wb_ack;
    //wb_dat <= 16'hxxxx;
    //feed_en <= 1'b0;
    case (wb_adr_i[7:0]) 
    8'h00: wb_dat <= 16'h4800; // Model ID
    8'h02: wb_dat <= {reset_fpga, strap_low, sd_power, boot_mode_i, bbclk, 1'b0, irq_can1, ext_temp_i, p_submodel, p_pldrev};
    8'h04: wb_dat <= timer_i[15:0];
    8'h06: wb_dat <= timer_i[31:16];
    8'h08: wb_dat <= usec[15:0];
    8'h0a: wb_dat <= usec[31:16];
    8'h0c: wb_dat <= random;
    8'h0e: begin
      //feed_en <= wb_cyc_i && wb_stb_i && wb_we_i && !feed_bsy;
      wb_ack <= wb_cyc_i && wb_stb_i && ((wb_we_i & !feed_bsy) || !wb_we_i);
    end
    8'h10: wb_dat <= core_enable;
    8'h12: wb_dat <= bus_config;
    8'h14: wb_dat <= {12'h000, spi2_clk, spi2_si, spi2_csn, spi_so};
    8'h16: wb_dat <= {counter_en, en_rs422, en_ts8510_rs422, 1'h0, pllphase, mode3_i, resetsw_en, scratch, mode2_i, mode1_i};
	8'h1a: wb_dat <= counter0;
	8'h1c: wb_dat <= counter1;	
    8'h1e: wb_dat <= test_reg[15:0];
    8'h20: wb_dat <= dio_i[15:0];
    8'h22: wb_dat <= dio[15:0];
    8'h24: wb_dat <= dio_oe[15:0];
    8'h28: wb_dat <= dio_i[31:16];
    8'h2a: wb_dat <= dio[31:16];
    8'h2c: wb_dat <= dio_oe[31:16];
    8'h30: wb_dat <= dio_i[47:32];
    8'h32: wb_dat <= dio[47:32];
    8'h34: wb_dat <= dio_oe[47:32];
    8'h38: wb_dat <= {8'h00, dio_i[55:48]};
    8'h3a: wb_dat <= {8'h00, dio[55:48]};
    8'h3c: wb_dat <= {8'h00, dio_oe[55:48]};
    8'ha0: wb_dat <= 16'hbeef;
    8'ha2: wb_dat <= 16'hdead;
    endcase
    if (!(wb_stb_i && wb_cyc_i)) wb_dat <= 16'h0000;
  end
end

always @(*) begin
    feed_en = 1'b0;
    case (wb_adr_i[7:0]) 
    8'h0e: feed_en = wb_cyc_i && wb_stb_i && wb_we_i && !feed_bsy;
    endcase
end


endmodule

